import json

import ipykernel
import socket
import os
from pyspark.sql import SparkSession
from pyspark.conf import SparkConf
# rest_utils.py comes from the base-image/python-feature-pipeline
from rest_utils import send_request, _parse_rest_error, HTTP_CONFIG, REST_CONFIG, DELIMITERS, ENV_VARIABLES


class SparkSessionInitializer(object):
    def __init__(self, ip):
        self.shell = ip
        self.spark = None
        connection_file = os.path.basename(ipykernel.get_connection_file())
        if 'kernel' in connection_file:
            self.kernel_id = connection_file.split('-', 1)[1].split('.')[0]

    def pre_execute(self):
        if not self.spark:
            print("Initializing spark session...")
            driver_port = get_free_port()
            conf = SparkConf()
            conf.set("spark.ui.port", driver_port)
            conf.set("spark.ui.proxyBase", "/hopsworks-api/sparkdriverui/" + \
                     os.environ["SPARK_JUPYTER_HEADLESS_SERVICE_NAME"] + ":" + \
                     str(driver_port) + "/project/" + os.environ["HOPSWORKS_PROJECT_ID"])
            conf.set("spark.ui.proxyRedirectUri", os.environ[ENV_VARIABLES.REST_ENDPOINT_END_VAR])
            self.spark = SparkSession.builder.config(conf=conf).enableHiveSupport().getOrCreate()
            self.shell.push({'spark': self.spark})
            print("Spark session available as `spark`.")
            try:
                send_spark_session_details(self.kernel_id, driver_port, self.spark.sparkContext.applicationId)
            except Exception as e:
                print("Error sending spark session details to Hopsworks: " + str(e) + " You will not be able to see "
                                                                                      "the Spark UI in Hopsworks.")


def load_ipython_extension(ip):
    spark_session_initializer = SparkSessionInitializer(ip)
    ip.events.register('pre_execute', spark_session_initializer.pre_execute)


def get_free_port():
    host = 'localhost'
    port = 4040
    maxAttempts = 60
    while maxAttempts >= 0:
        if is_port_free(host, port):
            return port
        port += 1
        maxAttempts -= 1
    raise Exception("Could not find a free port in the range 4040-4100")


def is_port_free(host, port):
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(2)  # Set a timeout for connection attempt
        s.connect((host, port))
        s.close()
        return False
    except socket.error:
        return True


def send_spark_session_details(kernel_id, port, application_id):
    method = HTTP_CONFIG.HTTP_PUT
    resource_url = DELIMITERS.SLASH_DELIMITER + \
                   REST_CONFIG.HOPSWORKS_REST_RESOURCE + DELIMITERS.SLASH_DELIMITER + \
                   "project" + DELIMITERS.SLASH_DELIMITER + os.environ["HOPSWORKS_PROJECT_ID"] + \
                   DELIMITERS.SLASH_DELIMITER + "jupyter/sparkSession" + \
                   DELIMITERS.SLASH_DELIMITER + os.environ["HADOOP_USER_NAME"]
    headers = {HTTP_CONFIG.HTTP_CONTENT_TYPE: HTTP_CONFIG.HTTP_APPLICATION_JSON}
    data = {'kernelId': kernel_id, 'port': int(port), 'applicationId': application_id}
    response = send_request(method, resource_url, data=json.dumps(data), headers=headers)
    if response.status_code >= 400:
        response_object = response.json()
        error_code, error_msg, user_msg = _parse_rest_error(response_object)
        print("Error sending spark session details to hopsworks: " + user_msg)
